@@ -1,4 +1,4 @@
-# Copyright 2015 DataStax, Inc.
+# Copyright 2013-2016 DataStax, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -13,16 +13,15 @@
 # limitations under the License.
 
 from datetime import datetime, timedelta
-import logging
 import time
 import six
-import warnings
+from six.moves import filter
 
+from cassandra.query import FETCH_SIZE_UNSET
+from cassandra.cqlengine import columns
 from cassandra.cqlengine import UnicodeMixin
 from cassandra.cqlengine.functions import QueryValue
-from cassandra.cqlengine.operators import BaseWhereOperator, InOperator
-
-log = logging.getLogger(__name__)
+from cassandra.cqlengine.operators import BaseWhereOperator, InOperator, EqualsOperator
 
 
 class StatementException(Exception):
@@ -108,7 +107,7 @@ class WhereClause(BaseClause):
         """
         if not isinstance(operator, BaseWhereOperator):
             raise StatementException(
-                "operator must be of type {}, got {}".format(BaseWhereOperator, type(operator))
+                "operator must be of type {0}, got {1}".format(BaseWhereOperator, type(operator))
             )
         super(WhereClause, self).__init__(field, value)
         self.operator = operator
@@ -116,8 +115,8 @@ class WhereClause(BaseClause):
         self.quote_field = quote_field
 
     def __unicode__(self):
-        field = ('"{}"' if self.quote_field else '{}').format(self.field)
-        return u'{} {} {}'.format(field, self.operator, six.text_type(self.query_value))
+        field = ('"{0}"' if self.quote_field else '{0}').format(self.field)
+        return u'{0} {1} {2}'.format(field, self.operator, six.text_type(self.query_value))
 
     def __hash__(self):
         return super(WhereClause, self).__hash__() ^ hash(self.operator)
@@ -145,34 +144,41 @@ class AssignmentClause(BaseClause):
     """ a single variable st statement """
 
     def __unicode__(self):
-        return u'"{}" = %({})s'.format(self.field, self.context_id)
+        return u'"{0}" = %({1})s'.format(self.field, self.context_id)
 
     def insert_tuple(self):
         return self.field, self.context_id
 
 
-class TransactionClause(BaseClause):
+class ConditionalClause(BaseClause):
     """ A single variable iff statement """
 
     def __unicode__(self):
-        return u'"{}" = %({})s'.format(self.field, self.context_id)
+        return u'"{0}" = %({1})s'.format(self.field, self.context_id)
 
     def insert_tuple(self):
         return self.field, self.context_id
 
 
+class ContainerUpdateTypeMapMeta(type):
+
+    def __init__(cls, name, bases, dct):
+        if not hasattr(cls, 'type_map'):
+            cls.type_map = {}
+        else:
+            cls.type_map[cls.col_type] = cls
+        super(ContainerUpdateTypeMapMeta, cls).__init__(name, bases, dct)
+
+
+@six.add_metaclass(ContainerUpdateTypeMapMeta)
 class ContainerUpdateClause(AssignmentClause):
 
-    def __init__(self, field, value, operation=None, previous=None, column=None):
+    def __init__(self, field, value, operation=None, previous=None):
         super(ContainerUpdateClause, self).__init__(field, value)
         self.previous = previous
         self._assignments = None
         self._operation = operation
         self._analyzed = False
-        self._column = column
-
-    def _to_database(self, val):
-        return self._column.to_database(val) if self._column else val
 
     def _analyze(self):
         raise NotImplementedError
@@ -187,10 +193,10 @@ class ContainerUpdateClause(AssignmentClause):
 class SetUpdateClause(ContainerUpdateClause):
     """ updates a set collection """
 
-    def __init__(self, field, value, operation=None, previous=None, column=None):
-        super(SetUpdateClause, self).__init__(field, value, operation, previous, column=column)
-        self._additions = None
-        self._removals = None
+    col_type = columns.Set
+
+    _additions = None
+    _removals = None
 
     def __unicode__(self):
         qs = []
@@ -199,9 +205,9 @@ class SetUpdateClause(ContainerUpdateClause):
                 self._assignments is None and
                 self._additions is None and
                 self._removals is None):
-            qs += ['"{}" = %({})s'.format(self.field, ctx_id)]
+            qs += ['"{0}" = %({1})s'.format(self.field, ctx_id)]
         if self._assignments is not None:
-            qs += ['"{}" = %({})s'.format(self.field, ctx_id)]
+            qs += ['"{0}" = %({1})s'.format(self.field, ctx_id)]
             ctx_id += 1
         if self._additions is not None:
             qs += ['"{0}" = "{0}" + %({1})s'.format(self.field, ctx_id)]
@@ -245,24 +251,24 @@ class SetUpdateClause(ContainerUpdateClause):
                 self._assignments is None and
                 self._additions is None and
                 self._removals is None):
-            ctx[str(ctx_id)] = self._to_database({})
+            ctx[str(ctx_id)] = set()
         if self._assignments is not None:
-            ctx[str(ctx_id)] = self._to_database(self._assignments)
+            ctx[str(ctx_id)] = self._assignments
             ctx_id += 1
         if self._additions is not None:
-            ctx[str(ctx_id)] = self._to_database(self._additions)
+            ctx[str(ctx_id)] = self._additions
             ctx_id += 1
         if self._removals is not None:
-            ctx[str(ctx_id)] = self._to_database(self._removals)
+            ctx[str(ctx_id)] = self._removals
 
 
 class ListUpdateClause(ContainerUpdateClause):
     """ updates a list collection """
 
-    def __init__(self, field, value, operation=None, previous=None, column=None):
-        super(ListUpdateClause, self).__init__(field, value, operation, previous, column=column)
-        self._append = None
-        self._prepend = None
+    col_type = columns.List
+
+    _append = None
+    _prepend = None
 
     def __unicode__(self):
         if not self._analyzed:
@@ -270,7 +276,7 @@ class ListUpdateClause(ContainerUpdateClause):
         qs = []
         ctx_id = self.context_id
         if self._assignments is not None:
-            qs += ['"{}" = %({})s'.format(self.field, ctx_id)]
+            qs += ['"{0}" = %({1})s'.format(self.field, ctx_id)]
             ctx_id += 1
 
         if self._prepend is not None:
@@ -292,18 +298,13 @@ class ListUpdateClause(ContainerUpdateClause):
             self._analyze()
         ctx_id = self.context_id
         if self._assignments is not None:
-            ctx[str(ctx_id)] = self._to_database(self._assignments)
+            ctx[str(ctx_id)] = self._assignments
             ctx_id += 1
         if self._prepend is not None:
-            msg = "Previous versions of cqlengine implicitly reversed prepended lists to account for CASSANDRA-8733. " \
-                  "THIS VERSION DOES NOT. This warning will be removed in a future release."
-            warnings.warn(msg)
-            log.warning(msg)
-
-            ctx[str(ctx_id)] = self._to_database(self._prepend)
+            ctx[str(ctx_id)] = self._prepend
             ctx_id += 1
         if self._append is not None:
-            ctx[str(ctx_id)] = self._to_database(self._append)
+            ctx[str(ctx_id)] = self._append
 
     def _analyze(self):
         """ works out the updates to be performed """
@@ -357,9 +358,9 @@ class ListUpdateClause(ContainerUpdateClause):
 class MapUpdateClause(ContainerUpdateClause):
     """ updates a map collection """
 
-    def __init__(self, field, value, operation=None, previous=None, column=None):
-        super(MapUpdateClause, self).__init__(field, value, operation, previous, column=column)
-        self._updates = None
+    col_type = columns.Map
+
+    _updates = None
 
     def _analyze(self):
         if self._operation == "update":
@@ -372,52 +373,54 @@ class MapUpdateClause(ContainerUpdateClause):
         self._analyzed = True
 
     def get_context_size(self):
-        if not self._analyzed:
-            self._analyze()
-        if self.previous is None and not self._updates:
+        if self.is_assignment:
             return 1
         return len(self._updates or []) * 2
 
     def update_context(self, ctx):
-        if not self._analyzed:
-            self._analyze()
         ctx_id = self.context_id
-        if self.previous is None and not self._updates:
+        if self.is_assignment:
             ctx[str(ctx_id)] = {}
         else:
             for key in self._updates or []:
                 val = self.value.get(key)
-                ctx[str(ctx_id)] = self._column.key_col.to_database(key) if self._column else key
-                ctx[str(ctx_id + 1)] = self._column.value_col.to_database(val) if self._column else val
+                ctx[str(ctx_id)] = key
+                ctx[str(ctx_id + 1)] = val
                 ctx_id += 2
 
-    def __unicode__(self):
+    @property
+    def is_assignment(self):
         if not self._analyzed:
             self._analyze()
+        return self.previous is None and not self._updates
+
+    def __unicode__(self):
         qs = []
 
         ctx_id = self.context_id
-        if self.previous is None and not self._updates:
-            qs += ['"{}" = %({})s'.format(self.field, ctx_id)]
+        if self.is_assignment:
+            qs += ['"{0}" = %({1})s'.format(self.field, ctx_id)]
         else:
             for _ in self._updates or []:
-                qs += ['"{}"[%({})s] = %({})s'.format(self.field, ctx_id, ctx_id + 1)]
+                qs += ['"{0}"[%({1})s] = %({2})s'.format(self.field, ctx_id, ctx_id + 1)]
                 ctx_id += 2
 
         return ', '.join(qs)
 
 
-class CounterUpdateClause(ContainerUpdateClause):
+class CounterUpdateClause(AssignmentClause):
 
-    def __init__(self, field, value, previous=None, column=None):
-        super(CounterUpdateClause, self).__init__(field, value, previous=previous, column=column)
-        self.previous = self.previous or 0
+    col_type = columns.Counter
+
+    def __init__(self, field, value, previous=None):
+        super(CounterUpdateClause, self).__init__(field, value)
+        self.previous = previous or 0
 
     def get_context_size(self):
         return 1
 
     def update_context(self, ctx):
-        ctx[str(self.context_id)] = self._to_database(abs(self.value - self.previous))
+        ctx[str(self.context_id)] = abs(self.value - self.previous)
 
     def __unicode__(self):
         delta = self.value - self.previous
@@ -436,7 +439,7 @@ class FieldDeleteClause(BaseDeleteClause):
         super(FieldDeleteClause, self).__init__(field, None)
 
     def __unicode__(self):
-        return '"{}"'.format(self.field)
+        return '"{0}"'.format(self.field)
 
     def update_context(self, ctx):
         pass
@@ -473,32 +476,43 @@ class MapDeleteClause(BaseDeleteClause):
     def __unicode__(self):
         if not self._analyzed:
             self._analyze()
-        return ', '.join(['"{}"[%({})s]'.format(self.field, self.context_id + i) for i in range(len(self._removals))])
+        return ', '.join(['"{0}"[%({1})s]'.format(self.field, self.context_id + i) for i in range(len(self._removals))])
 
 
 class BaseCQLStatement(UnicodeMixin):
     """ The base cql statement class """
 
-    def __init__(self, table, consistency=None, timestamp=None, where=None):
+    def __init__(self, table, timestamp=None, where=None, fetch_size=None, conditionals=None):
         super(BaseCQLStatement, self).__init__()
         self.table = table
-        self.consistency = consistency
         self.context_id = 0
         self.context_counter = self.context_id
         self.timestamp = timestamp
+        self.fetch_size = fetch_size if fetch_size else FETCH_SIZE_UNSET
 
         self.where_clauses = []
         for clause in where or []:
-            self.add_where_clause(clause)
+            self._add_where_clause(clause)
 
-    def add_where_clause(self, clause):
-        """
-        adds a where clause to this statement
-        :param clause: the clause to add
-        :type clause: WhereClause
-        """
-        if not isinstance(clause, WhereClause):
-            raise StatementException("only instances of WhereClause can be added to statements")
+        self.conditionals = []
+        for conditional in conditionals or []:
+            self.add_conditional_clause(conditional)
+
+    def _update_part_key_values(self, field_index_map, clauses, parts):
+        for clause in filter(lambda c: c.field in field_index_map, clauses):
+            parts[field_index_map[clause.field]] = clause.value
+
+    def partition_key_values(self, field_index_map):
+        parts = [None] * len(field_index_map)
+        self._update_part_key_values(field_index_map, (w for w in self.where_clauses if w.operator.__class__ == EqualsOperator), parts)
+        return parts
+
+    def add_where(self, column, operator, value, quote_field=True):
+        value = column.to_database(value)
+        clause = WhereClause(column.db_field_name, operator, value, quote_field)
+        self._add_where_clause(clause)
+
+    def _add_where_clause(self, clause):
         clause.set_context_id(self.context_counter)
         self.context_counter += clause.get_context_size()
         self.where_clauses.append(clause)
@@ -513,6 +527,20 @@ class BaseCQLStatement(UnicodeMixin):
             clause.update_context(ctx)
         return ctx
 
+    def add_conditional_clause(self, clause):
+        """
+        Adds a iff clause to this statement
+
+        :param clause: The clause that will be added to the iff statement
+        :type clause: ConditionalClause
+        """
+        clause.set_context_id(self.context_counter)
+        self.context_counter += clause.get_context_size()
+        self.conditionals.append(clause)
+
+    def _get_conditionals(self):
+        return 'IF {0}'.format(' AND '.join([six.text_type(c) for c in self.conditionals]))
+
     def get_context_size(self):
         return len(self.get_context())
 
@@ -550,7 +578,7 @@ class BaseCQLStatement(UnicodeMixin):
 
     @property
     def _where(self):
-        return 'WHERE {}'.format(' AND '.join([six.text_type(c) for c in self.where_clauses]))
+        return 'WHERE {0}'.format(' AND '.join([six.text_type(c) for c in self.where_clauses]))
 
 
 class SelectStatement(BaseCQLStatement):
@@ -560,11 +588,12 @@ class SelectStatement(BaseCQLStatement):
                  table,
                  fields=None,
                  count=False,
-                 consistency=None,
                  where=None,
                  order_by=None,
                  limit=None,
-                 allow_filtering=False):
+                 allow_filtering=False,
+                 distinct_fields=None,
+                 fetch_size=None):
 
         """
         :param where
@@ -572,11 +601,12 @@ class SelectStatement(BaseCQLStatement):
         """
         super(SelectStatement, self).__init__(
             table,
-            consistency=consistency,
-            where=where
+            where=where,
+            fetch_size=fetch_size
         )
 
         self.fields = [fields] if isinstance(fields, six.string_types) else (fields or [])
+        self.distinct_fields = distinct_fields
         self.count = count
         self.order_by = [order_by] if isinstance(order_by, six.string_types) else order_by
         self.limit = limit
@@ -584,20 +614,25 @@ class SelectStatement(BaseCQLStatement):
 
     def __unicode__(self):
         qs = ['SELECT']
-        if self.count:
+        if self.distinct_fields:
+            if self.count:
+                qs += ['DISTINCT COUNT({0})'.format(', '.join(['"{0}"'.format(f) for f in self.distinct_fields]))]
+            else:
+                qs += ['DISTINCT {0}'.format(', '.join(['"{0}"'.format(f) for f in self.distinct_fields]))]
+        elif self.count:
             qs += ['COUNT(*)']
         else:
-            qs += [', '.join(['"{}"'.format(f) for f in self.fields]) if self.fields else '*']
+            qs += [', '.join(['"{0}"'.format(f) for f in self.fields]) if self.fields else '*']
         qs += ['FROM', self.table]
 
         if self.where_clauses:
             qs += [self._where]
 
         if self.order_by and not self.count:
-            qs += ['ORDER BY {}'.format(', '.join(six.text_type(o) for o in self.order_by))]
+            qs += ['ORDER BY {0}'.format(', '.join(six.text_type(o) for o in self.order_by))]
 
         if self.limit:
-            qs += ['LIMIT {}'.format(self.limit)]
+            qs += ['LIMIT {0}'.format(self.limit)]
 
         if self.allow_filtering:
             qs += ['ALLOW FILTERING']
@@ -611,14 +646,14 @@ class AssignmentStatement(BaseCQLStatement):
     def __init__(self,
                  table,
                  assignments=None,
-                 consistency=None,
                  where=None,
                  ttl=None,
-                 timestamp=None):
+                 timestamp=None,
+                 conditionals=None):
         super(AssignmentStatement, self).__init__(
             table,
-            consistency=consistency,
             where=where,
+            conditionals=conditionals
         )
         self.ttl = ttl
         self.timestamp = timestamp
@@ -626,7 +661,7 @@ class AssignmentStatement(BaseCQLStatement):
         # add assignments
         self.assignments = []
         for assignment in assignments or []:
-            self.add_assignment_clause(assignment)
+            self._add_assignment_clause(assignment)
 
     def update_context_id(self, i):
         super(AssignmentStatement, self).update_context_id(i)
@@ -634,14 +669,17 @@ class AssignmentStatement(BaseCQLStatement):
             assignment.set_context_id(self.context_counter)
             self.context_counter += assignment.get_context_size()
 
-    def add_assignment_clause(self, clause):
-        """
-        adds an assignment clause to this statement
-        :param clause: the clause to add
-        :type clause: AssignmentClause
-        """
-        if not isinstance(clause, AssignmentClause):
-            raise StatementException("only instances of AssignmentClause can be added to statements")
+    def partition_key_values(self, field_index_map):
+        parts = super(AssignmentStatement, self).partition_key_values(field_index_map)
+        self._update_part_key_values(field_index_map, self.assignments, parts)
+        return parts
+
+    def add_assignment(self, column, value):
+        value = column.to_database(value)
+        clause = AssignmentClause(column.db_field_name, value)
+        self._add_assignment_clause(clause)
+
+    def _add_assignment_clause(self, clause):
         clause.set_context_id(self.context_counter)
         self.context_counter += clause.get_context_size()
         self.assignments.append(clause)
@@ -658,47 +696,42 @@ class AssignmentStatement(BaseCQLStatement):
 
 
 class InsertStatement(AssignmentStatement):
-    """ an cql insert select statement """
+    """ an cql insert statement """
 
     def __init__(self,
                  table,
                  assignments=None,
-                 consistency=None,
                  where=None,
                  ttl=None,
                  timestamp=None,
                  if_not_exists=False):
         super(InsertStatement, self).__init__(table,
                                               assignments=assignments,
-                                              consistency=consistency,
                                               where=where,
                                               ttl=ttl,
                                               timestamp=timestamp)
 
         self.if_not_exists = if_not_exists
 
-    def add_where_clause(self, clause):
-        raise StatementException("Cannot add where clauses to insert statements")
-
     def __unicode__(self):
-        qs = ['INSERT INTO {}'.format(self.table)]
+        qs = ['INSERT INTO {0}'.format(self.table)]
 
         # get column names and context placeholders
         fields = [a.insert_tuple() for a in self.assignments]
         columns, values = zip(*fields)
 
-        qs += ["({})".format(', '.join(['"{}"'.format(c) for c in columns]))]
+        qs += ["({0})".format(', '.join(['"{0}"'.format(c) for c in columns]))]
         qs += ['VALUES']
-        qs += ["({})".format(', '.join(['%({})s'.format(v) for v in values]))]
+        qs += ["({0})".format(', '.join(['%({0})s'.format(v) for v in values]))]
 
         if self.if_not_exists:
             qs += ["IF NOT EXISTS"]
 
         if self.ttl:
-            qs += ["USING TTL {}".format(self.ttl)]
+            qs += ["USING TTL {0}".format(self.ttl)]
 
         if self.timestamp:
-            qs += ["USING TIMESTAMP {}".format(self.timestamp_normalized)]
+            qs += ["USING TIMESTAMP {0}".format(self.timestamp_normalized)]
 
         return ' '.join(qs)
 
@@ -709,22 +742,19 @@ class UpdateStatement(AssignmentStatement):
     def __init__(self,
                  table,
                  assignments=None,
-                 consistency=None,
                  where=None,
                  ttl=None,
                  timestamp=None,
-                 transactions=None):
+                 conditionals=None,
+                 if_exists=False):
         super(UpdateStatement, self). __init__(table,
                                                assignments=assignments,
-                                               consistency=consistency,
                                                where=where,
                                                ttl=ttl,
-                                               timestamp=timestamp)
+                                               timestamp=timestamp,
+                                               conditionals=conditionals)
 
-        # Add iff statements
-        self.transactions = []
-        for transaction in transactions or []:
-            self.add_transaction_clause(transaction)
+        self.if_exists = if_exists
 
     def __unicode__(self):
         qs = ['UPDATE', self.table]
@@ -732,13 +762,13 @@ class UpdateStatement(AssignmentStatement):
         using_options = []
 
         if self.ttl:
-            using_options += ["TTL {}".format(self.ttl)]
+            using_options += ["TTL {0}".format(self.ttl)]
 
         if self.timestamp:
-            using_options += ["TIMESTAMP {}".format(self.timestamp_normalized)]
+            using_options += ["TIMESTAMP {0}".format(self.timestamp_normalized)]
 
         if using_options:
-            qs += ["USING {}".format(" AND ".join(using_options))]
+            qs += ["USING {0}".format(" AND ".join(using_options))]
 
         qs += ['SET']
         qs += [', '.join([six.text_type(c) for c in self.assignments])]
@@ -746,49 +776,50 @@ class UpdateStatement(AssignmentStatement):
         if self.where_clauses:
             qs += [self._where]
 
-        if len(self.transactions) > 0:
-            qs += [self._get_transactions()]
+        if len(self.conditionals) > 0:
+            qs += [self._get_conditionals()]
+
+        if self.if_exists:
+            qs += ["IF EXISTS"]
 
         return ' '.join(qs)
 
-    def add_transaction_clause(self, clause):
-        """
-        Adds a iff clause to this statement
-
-        :param clause: The clause that will be added to the iff statement
-        :type clause: TransactionClause
-        """
-        if not isinstance(clause, TransactionClause):
-            raise StatementException('only instances of AssignmentClause can be added to statements')
-        clause.set_context_id(self.context_counter)
-        self.context_counter += clause.get_context_size()
-        self.transactions.append(clause)
-
     def get_context(self):
         ctx = super(UpdateStatement, self).get_context()
-        for clause in self.transactions or []:
+        for clause in self.conditionals:
             clause.update_context(ctx)
         return ctx
 
-    def _get_transactions(self):
-        return 'IF {}'.format(' AND '.join([six.text_type(c) for c in self.transactions]))
-
     def update_context_id(self, i):
         super(UpdateStatement, self).update_context_id(i)
-        for transaction in self.transactions:
-            transaction.set_context_id(self.context_counter)
-            self.context_counter += transaction.get_context_size()
+        for conditional in self.conditionals:
+            conditional.set_context_id(self.context_counter)
+            self.context_counter += conditional.get_context_size()
+
+    def add_update(self, column, value, operation=None, previous=None):
+        value = column.to_database(value)
+        col_type = type(column)
+        container_update_type = ContainerUpdateClause.type_map.get(col_type)
+        if container_update_type:
+            previous = column.to_database(previous)
+            clause = container_update_type(column.db_field_name, value, operation, previous)
+        elif col_type == columns.Counter:
+            clause = CounterUpdateClause(column.db_field_name, value, previous)
+        else:
+            clause = AssignmentClause(column.db_field_name, value)
+        if clause.get_context_size():  # this is to exclude map removals from updates. Can go away if we drop support for C* < 1.2.4 and remove two-phase updates
+            self._add_assignment_clause(clause)
 
 
 class DeleteStatement(BaseCQLStatement):
     """ a cql delete statement """
 
-    def __init__(self, table, fields=None, consistency=None, where=None, timestamp=None):
+    def __init__(self, table, fields=None, where=None, timestamp=None, conditionals=None, if_exists=False):
         super(DeleteStatement, self).__init__(
             table,
-            consistency=consistency,
             where=where,
-            timestamp=timestamp
+            timestamp=timestamp,
+            conditionals=conditionals
         )
         self.fields = []
         if isinstance(fields, six.string_types):
@@ -796,16 +827,23 @@ class DeleteStatement(BaseCQLStatement):
         for field in fields or []:
             self.add_field(field)
 
+        self.if_exists = if_exists
+
     def update_context_id(self, i):
         super(DeleteStatement, self).update_context_id(i)
         for field in self.fields:
             field.set_context_id(self.context_counter)
             self.context_counter += field.get_context_size()
+        for t in self.conditionals:
+            t.set_context_id(self.context_counter)
+            self.context_counter += t.get_context_size()
 
     def get_context(self):
         ctx = super(DeleteStatement, self).get_context()
         for field in self.fields:
             field.update_context(ctx)
+        for clause in self.conditionals:
+            clause.update_context(ctx)
         return ctx
 
     def add_field(self, field):
@@ -820,18 +858,24 @@ class DeleteStatement(BaseCQLStatement):
     def __unicode__(self):
         qs = ['DELETE']
         if self.fields:
-            qs += [', '.join(['{}'.format(f) for f in self.fields])]
+            qs += [', '.join(['{0}'.format(f) for f in self.fields])]
         qs += ['FROM', self.table]
 
         delete_option = []
 
         if self.timestamp:
-            delete_option += ["TIMESTAMP {}".format(self.timestamp_normalized)]
+            delete_option += ["TIMESTAMP {0}".format(self.timestamp_normalized)]
 
         if delete_option:
-            qs += [" USING {} ".format(" AND ".join(delete_option))]
+            qs += [" USING {0} ".format(" AND ".join(delete_option))]
 
         if self.where_clauses:
             qs += [self._where]
 
+        if self.conditionals:
+            qs += [self._get_conditionals()]
+
+        if self.if_exists:
+            qs += ["IF EXISTS"]
+
         return ' '.join(qs)
